"""Resolwe entity serializer."""
from django.db import transaction

from resolwe.flow.models import Collection, Entity
from resolwe.permissions.models import Permission

from .collection import BaseCollectionSerializer, CollectionSerializer
from .fields import DictRelatedField


class BaseEntitySerializer(BaseCollectionSerializer):
    """Entity serializer without the collection field.

    Used when serializing data object, since collection info is duplicated.
    """

    class Meta(BaseCollectionSerializer.Meta):
        """EntitySerializer Meta options."""

        model = Entity
        fields = BaseCollectionSerializer.Meta.fields + (
            "collection",
            "duplicated",
            "type",
        )

    @transaction.atomic
    def update(self, instance, validated_data):
        """Update collection."""
        update_collection = "collection" in validated_data
        new_collection = validated_data.pop("collection", None)
        instance = super().update(instance, validated_data)
        if update_collection and new_collection != instance.collection:
            instance.move_to_collection(new_collection)
        return instance


class EntitySerializer(BaseEntitySerializer):
    """Serializer for Entity with added collection field."""

    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=CollectionSerializer,
        allow_null=True,
        required=False,
        write_permission=Permission.EDIT,
    )



# Generated by Django 3.2.15 on 2022-08-26 09:02

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import resolwe.observers.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Observer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField(null=True)),
                (
                    "change_type",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "CREATE"), (2, "UPDATE"), (3, "DELETE")]
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("session_id", models.CharField(max_length=100)),
                (
                    "subscription_id",
                    models.UUIDField(
                        default=resolwe.observers.models.get_random_uuid,
                        editable=False,
                        unique=True,
                    ),
                ),
                (
                    "observers",
                    models.ManyToManyField(
                        related_name="subscriptions", to="observers.Observer"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="subscription",
            index=models.Index(
                fields=["session_id"], name="observers_s_session_d8c2db_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="observer",
            index=models.Index(
                fields=["object_id"], name="observers_o_object__f89e17_idx"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="observer",
            unique_together={("content_type", "object_id", "change_type")},
        ),
    ]



# pylint: disable=missing-docstring,invalid-name
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.urls import reverse

from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, Process
from resolwe.flow.views import CollectionViewSet, StorageViewSet
from resolwe.permissions.models import Permission, get_anonymous_user
from resolwe.permissions.shortcuts import get_user_group_perms
from resolwe.test import TestCase

factory = APIRequestFactory()


class UserGroupTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")

        self.collection = Collection.objects.create(
            contributor=self.contributor,
            name="Test collection",
        )

        # This collection is here to make sure that other permissions
        # don't affect tested queries.
        collection2 = Collection.objects.create(
            contributor=self.contributor,
            name="Test collection 2",
        )
        collection2.set_permission(Permission.VIEW, self.contributor)
        collection2.set_permission(Permission.VIEW, self.group1)

    def test_user(self):
        self.collection.set_permission(Permission.EDIT, self.contributor)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )

        self.assertEqual(len(group_perms), 0)
        self.assertCountEqual(user_perms, [Permission.VIEW, Permission.EDIT])

    def test_user_in_group(self):
        self.group1.user_set.add(self.contributor)
        self.collection.set_permission(Permission.EDIT, self.group1)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], [Permission.VIEW, Permission.EDIT])
        self.assertEqual(len(user_perms), 0)

        self.collection.set_permission(Permission.VIEW, self.contributor)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], [Permission.VIEW, Permission.EDIT])
        self.assertEqual(len(user_perms), 1)
        self.assertCountEqual(user_perms, [Permission.VIEW])

    def test_user_in_multiple_groups(self):
        self.group1.user_set.add(self.contributor)
        self.group2.user_set.add(self.contributor)
        self.collection.set_permission(Permission.EDIT, self.group1)
        self.collection.set_permission(Permission.VIEW, self.group2)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 2)
        self.assertEqual(group_perms[0][0], self.group1.pk)
        self.assertCountEqual(group_perms[0][2], [Permission.VIEW, Permission.EDIT])
        self.assertEqual(group_perms[1][0], self.group2.pk)
        self.assertCountEqual(group_perms[1][2], [Permission.VIEW])
        self.assertEqual(len(user_perms), 0)

    def test_group(self):
        self.collection.set_permission(Permission.EDIT, self.group1)
        user_perms, group_perms = get_user_group_perms(self.group1, self.collection)
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], [Permission.VIEW, Permission.EDIT])
        self.assertEqual(len(user_perms), 0)


class ObjectPermsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.admin.delete()
        self.user1 = get_user_model().objects.create(username="test_user1")
        self.user2 = get_user_model().objects.create(username="test_user2")
        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")
        self.anonymous = get_anonymous_user()

        self.collection = Collection.objects.create(
            contributor=self.user1,
            name="Test collection",
        )

    def _sort_perms(self, perms):
        for elm in perms:
            elm["permissions"] = sorted(elm["permissions"])
        return perms

    def test_all_permissions(self):
        """Test all user permissions.

        Method get_object_perms should only be tested through view, as it
        expects pre-fetched permissions for the given user.

        The request to get all users permissions must be made by the user with
        permission level SHARE or higher, otherwise only permissions for the
        current user will be returned.
        """
        self.group1.user_set.add(self.user1)
        self.collection.set_permission(Permission.EDIT, self.user1)
        self.collection.set_permission(Permission.SHARE, self.user2)

        url = reverse(
            "resolwe-api:collection-permissions", kwargs={"pk": self.collection.pk}
        )
        request = factory.get(url, data={}, format="json")
        force_authenticate(request, self.user2)
        detail_view = CollectionViewSet.as_view({"get": "detail_permissions"})

        expected_perms = [
            {
                "permissions": ["edit", "view"],
                "type": "user",
                "id": self.user1.pk,
                "name": "test_user1",
                "username": "test_user1",
            },
            {
                "permissions": ["edit", "share", "view"],
                "type": "user",
                "id": self.user2.pk,
                "name": "test_user2",
                "username": "test_user2",
            },
        ]

        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        self.collection.set_permission(Permission.EDIT, self.group1)
        self.collection.set_permission(Permission.VIEW, self.group2)
        expected_perms.extend(
            [
                {
                    "permissions": ["edit", "view"],
                    "type": "group",
                    "id": self.group1.pk,
                    "name": "Test group 1",
                },
                {
                    "permissions": ["view"],
                    "type": "group",
                    "id": self.group2.pk,
                    "name": "Test group 2",
                },
            ]
        )
        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        self.collection.set_permission(Permission.VIEW, self.anonymous)
        expected_perms.append(
            {"permissions": ["view"], "type": "public"},
        )
        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

    def test_user_permissions(self):
        """Test user permissions.

        Method get_object_perms should only be tested through view, as it
        expects pre-fetched permissions for the given user.
        """

        self.group1.user_set.add(self.user1)
        self.collection.set_permission(Permission.EDIT, self.user1)
        self.collection.set_permission(Permission.VIEW, self.user2)
        self.collection.set_permission(Permission.EDIT, self.group1)
        self.collection.set_permission(Permission.VIEW, self.group2)

        expected_perms = [
            {
                "permissions": ["edit", "view"],
                "type": "user",
                "id": self.user1.pk,
                "name": "test_user1",
                "username": "test_user1",
            },
            {
                "permissions": ["edit", "view"],
                "type": "group",
                "id": self.group1.pk,
                "name": "Test group 1",
            },
        ]

        url = reverse(
            "resolwe-api:collection-permissions", kwargs={"pk": self.collection.pk}
        )
        request = factory.get(url, data={}, format="json")
        force_authenticate(request, self.user1)
        detail_view = CollectionViewSet.as_view({"get": "detail_permissions"})
        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        self.group2.user_set.add(self.user1)
        expected_perms.append(
            {
                "permissions": ["view"],
                "type": "group",
                "id": self.group2.pk,
                "name": "Test group 2",
            },
        )

        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        self.collection.set_permission(Permission.VIEW, self.anonymous)
        expected_perms.append(
            {"permissions": ["view"], "type": "public"},
        )

        response = detail_view(request, pk=self.collection.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        perms = response.data
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

    def test_users_with_permission(self):
        self.group1.user_set.add(self.user1)
        self.collection.set_permission(Permission.EDIT, self.user1)
        self.collection.set_permission(Permission.SHARE, self.user2)
        admin = get_user_model().objects.create_superuser(
            username="admin",
            email="admin@test.com",
            password="admin",
            first_name="James",
            last_name="Smith",
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.VIEW),
            [self.user1, self.user2],
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.VIEW, True),
            [self.user1, self.user2, admin],
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.SHARE),
            [self.user2],
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.SHARE, True),
            [self.user2, admin],
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.OWNER),
            [],
        )
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.OWNER, True),
            [admin],
        )
        self.collection.set_permission(Permission.NONE, self.user1)
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.VIEW),
            [self.user2],
        )
        self.collection.set_permission(Permission.VIEW, self.group1)
        self.assertCountEqual(
            self.collection.users_with_permission(Permission.VIEW),
            [self.user1, self.user2],
        )


class StoragePermsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        proc = Process.objects.create(name="Test process", contributor=self.contributor)
        self.data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=proc
        )
        dummy_data = Data.objects.create(
            name="Dummy data", contributor=self.contributor, process=proc
        )

        self.storage1 = self.data.storages.create(
            name="Test storage",
            json={},
            contributor=self.contributor,
        )

        self.storage2 = self.data.storages.create(
            name="Test storage 2",
            json={},
            contributor=self.contributor,
        )

        dummy_data.storages.create(
            name="Dummy storage",
            json={},
            contributor=self.contributor,
        )

        self.user = get_user_model().objects.create(username="test_user")
        self.group = Group.objects.create(name="test_group")

        self.storage_list_viewset = StorageViewSet.as_view(
            actions={
                "get": "list",
            }
        )

        self.storage_detail_viewset = StorageViewSet.as_view(
            actions={
                "get": "retrieve",
            }
        )

    def test_detail_permissons(self):
        request = factory.get("/", content_type="application/json")
        force_authenticate(request, self.user)

        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        self.data.set_permission(Permission.VIEW, self.user)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.data["name"], "Test storage")

        self.data.set_permission(Permission.NONE, self.user)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

    def test_not_allowed_methods(self):
        self.assertFalse(hasattr(StorageViewSet, "update"))
        self.assertFalse(hasattr(StorageViewSet, "partial_update"))
        self.assertFalse(hasattr(StorageViewSet, "destroy"))
        self.assertFalse(hasattr(StorageViewSet, "create"))



"""Communication container startup script."""

import array
import asyncio
import functools
import json
import logging
import os
import shutil
import signal
import socket
import sys
import threading
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from distutils.util import strtobool
from pathlib import Path
from typing import Any, Optional, Tuple

import zmq
import zmq.asyncio
from executors import constants, global_settings
from executors.connectors import connectors
from executors.connectors.baseconnector import BaseStorageConnector
from executors.connectors.hasher import StreamHasher
from executors.connectors.transfer import Transfer
from executors.socket_utils import (
    BaseCommunicator,
    BaseProtocol,
    Message,
    PeerIdentity,
    Response,
    ResponseStatus,
    SocketCommunicator,
)
from executors.transfer import transfer_data
from executors.zeromq_utils import ZMQCommunicator

# Socket used to connect with the processing container.
PROCESSING_SOCKET = constants.SOCKETS_VOLUME / constants.COMMUNICATION_PROCESSING_SOCKET
UPLOAD_FILE_SOCKET = constants.SOCKETS_VOLUME / constants.UPLOAD_FILE_SOCKET

# Listener IP and port are read from environment.
LISTENER_IP = os.getenv("LISTENER_SERVICE_HOST", "127.0.0.1")
LISTENER_PORT = os.getenv("LISTENER_SERVICE_PORT", "53893")
LISTENER_PROTOCOL = os.getenv("LISTENER_PROTOCOL", "tcp")
DATA_ID = int(os.getenv("DATA_ID", "-1"))
KEEP_DATA = bool(strtobool(os.environ.get("FLOW_MANAGER_KEEP_DATA", "False")))
RUNNING_IN_KUBERNETES = bool(
    strtobool(os.environ.get("RUNNING_IN_KUBERNETES", "False"))
)

# How many file descriptors to receive over socket in a single message.
DESCRIPTOR_CHUNK_SIZE = int(os.environ.get("DESCRIPTOR_CHUNK_SIZE", 100))

MOUNTED_CONNECTORS = [
    name for name in os.environ["MOUNTED_CONNECTORS"].split(",") if name
]

# Mapping between storage and connectors for this storage.
# The values are tuples: (default connector, default mounted connector)
STORAGE_CONNECTOR: dict[
    str, Tuple[BaseStorageConnector, Optional[BaseStorageConnector]]
] = {}

# Configure container logger. All logs are output to stdout for further
# processing.
# The log level defaults to debug except for boto and google loggers, which
# is set to warning due to them being extremely verbose in the debug mode.
LOG_LEVEL = int(os.getenv("LOG_LEVEL", logging.DEBUG))
BOTO_LOG_LEVEL = int(os.getenv("BOTO_LOG_LEVEL", logging.WARNING))
GOOGLE_LOG_LEVEL = int(os.getenv("GOOGLE_LOG_LEVEL", logging.WARNING))

logging.basicConfig(
    stream=sys.stdout,
    level=LOG_LEVEL,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

for boto_logger_name in ["botocore", "boto3", "s3transfer", "urllib3"]:
    logging.getLogger(boto_logger_name).setLevel(BOTO_LOG_LEVEL)

for google_logger in ["google"]:
    logging.getLogger(google_logger).setLevel(GOOGLE_LOG_LEVEL)

logger = logging.getLogger(__name__)

logger.info("Starting communication container for data with id %d.", DATA_ID)


def purge_secrets():
    """Purge the content of the secrets directory.

    The directory itself is mounted as volume, so it can not be deleted.
    """

    def handle_error(func, path, exc_info):
        """Handle permission errors while removing data directories."""
        if isinstance(exc_info[1], PermissionError):
            os.chmod(path, 0o700)
            shutil.rmtree(path)

    try:
        for root, dirs, files in os.walk(constants.SECRETS_VOLUME):
            for f in files:
                os.chmod(os.path.join(root, f), 0o700)
                os.unlink(os.path.join(root, f))
            for d in dirs:
                shutil.rmtree(os.path.join(root, d), onerror=handle_error)

    except OSError:
        logger.exception("Manager exception while removing secrets directory.")


class FakeConnector(BaseStorageConnector):
    """Fake connector class.

    Used to send it to the Transfer class.
    """

    def __init__(
        self,
        config: dict,
        name: str,
        file_streams: dict[str, Any],
        hashes: dict[str, str],
    ):
        """Connector initialization."""
        super().__init__(config, name)
        self.path = Path()
        self.supported_hash = ["crc32c", "md5", "awss3etag"]
        self.multipart_chunksize = self.CHUNK_SIZE
        self.file_streams = file_streams
        self.hashes = hashes
        self.get_ensures_data_integrity = True

    def duplicate(self):
        """Return self."""
        return self

    def base_path(self):
        """Get the base path."""
        return self.path

    def get_object_list(self, url):
        """Get a list of objects stored bellow the given URL."""
        raise NotImplementedError

    def push(self, stream, url):
        """Push data from the stream to the given URL."""
        raise NotImplementedError

    def exists(self, url):
        """Get if the object at the given URL exist."""
        raise NotImplementedError

    def get_hash(self, url, hash_type):
        """Get the hash of the given type for the given object."""
        return self.hashes[os.fspath(url)][hash_type]

    def get_hashes(self, url, hash_types):
        """Get the hashes of the given types for the given object."""
        raise NotImplementedError

    def set_hashes(self, url, hashes):
        """Set the  hashes for the given object."""
        raise NotImplementedError

    def get(self, url, stream):
        """Get data from the given URL and write it into the given stream."""
        raise NotImplementedError

    @property
    def can_open_stream(self):
        """Get True if connector can open object as stream."""
        return True

    def presigned_url(
        self,
        url,
        expiration=10,
        force_download=False,
    ):
        """Create a presigned URL."""
        raise NotImplementedError

    def open_stream(self, url, mode):
        """Get stream for data at the given URL."""
        # Skip 'subpath/' in "subpath/...".
        path = Path(url)
        if global_settings.LOCATION_SUBPATH < path:
            url = path.relative_to(global_settings.LOCATION_SUBPATH)
        file_stream = self.file_streams[os.fspath(url)]
        return file_stream


class Uploader:
    """Upload referenced files to remote location."""

    def __init__(self, manager: "Manager", loop: asyncio.AbstractEventLoop):
        """Initialize."""
        super().__init__()
        self._terminating = False
        self.manager = manager
        self.loop = loop
        self.ready = threading.Event()

    def receive_file_descriptors(
        self, sock: socket.SocketType
    ) -> Tuple[str, dict[str, Any], bool]:
        """Receive file descriptors.

        See https://docs.python.org/3/library/socket.html#socket.socket.recvmsg .

        Protocol:
        1. Size of the filenames array (json) (64bites, #filenames_length).
        2. Connector name, filenames, presigned flag & file descriptors.
        3. Send response in the form {"success": bool, "presigned_urls": list}.
        """
        filenames_length = int.from_bytes(sock.recv(8), byteorder="big")
        if filenames_length == 0:
            return ("", dict(), False)
        fds = array.array("i")  # Array of ints
        msg, ancdata, flags, addr = sock.recvmsg(
            filenames_length, socket.CMSG_LEN(DESCRIPTOR_CHUNK_SIZE * fds.itemsize)
        )
        logger.debug("Received file descriptors for files: %s.", msg)
        storage_name, filenames, need_presigned_urls = json.loads(msg.decode())
        for cmsg_level, cmsg_type, cmsg_data in ancdata:
            if cmsg_level == socket.SOL_SOCKET and cmsg_type == socket.SCM_RIGHTS:
                # Append data, ignoring any truncated integers at the end.
                fds.frombytes(
                    cmsg_data[: len(cmsg_data) - (len(cmsg_data) % fds.itemsize)]
                )
        return (storage_name, dict(zip(filenames, fds)), need_presigned_urls)

    def send_message(self, sock, response):
        """Send message to the socket."""
        payload = json.dumps(response).encode()
        sock.sendall(len(payload).to_bytes(8, byteorder="big"))
        sock.sendall(payload)

    def run(self):
        """Start listening for file descriptors.

        Socket must be constructed here since it is not thread safe.
        """
        server_socket = socket.socket(family=socket.AF_UNIX)
        server_socket.settimeout(constants.CONTAINER_TIMEOUT)
        server_socket.bind(os.fspath(UPLOAD_FILE_SOCKET))
        server_socket.listen()
        self.ready.set()
        # Wait for the connection up to CONTAINER_TIMEOUT seconds.
        # If it fails, the socket will be closed and process terminated if
        # it will try to save some files as outputs.
        try:
            client, info = server_socket.accept()
        except socket.timeout:
            logger.error("Processing container is not connected to the upload socket.")
            server_socket.close()
            raise

        # Set the timeout for blocking operations to 1 second so we can
        # check for terminating condition.
        client.settimeout(1)
        logger.info("Uploader client connected (%s)." % client)
        with client:
            while not self._terminating:
                try:
                    (
                        storage_name,
                        file_descriptors,
                        need_presigned_urls,
                    ) = self.receive_file_descriptors(client)
                    # Stop if the socket has been closed.
                    if not file_descriptors:
                        break
                except socket.timeout:
                    continue
                except:
                    logger.exception(
                        "Exception while receiving file descriptors, exiting."
                    )
                    break
                try:
                    presigned_urls = []
                    to_transfer = []
                    logger.debug("Got %s", file_descriptors)
                    hashes: dict[str, str] = dict()
                    referenced_files: dict[str, dict[str, str]] = dict()

                    file_streams = {
                        file_name: os.fdopen(file_descriptor, "rb")
                        for file_name, file_descriptor in file_descriptors.items()
                    }

                    # Get default connector for the given storage name.
                    to_connector = STORAGE_CONNECTOR[storage_name][0]
                    for file_name in file_descriptors:
                        file_descriptor = file_descriptors[file_name]
                        stream = file_streams[file_name]
                        file_size = os.stat(file_descriptor).st_size

                        # Chose chunk size for S3. The chunk_size must be such
                        # that the file_size fits in at most 10_000 chunks. See
                        # https://docs.aws.amazon.com/AmazonS3/latest/dev/mpuoverview.html
                        # for additional information about this hard limit.

                        # Min chunk size must be 8 mega bytes. This is also the
                        # threshold for multipart uploads.
                        min_chunk_size = 8 * 1024 * 1024
                        needed_chunk_size = int(file_size / 10000) + 1
                        chunk_size = max(min_chunk_size, needed_chunk_size)

                        hasher = StreamHasher(chunk_size=chunk_size)
                        hasher.compute(stream)
                        referenced_files[file_name] = {
                            hash_type: hasher.hexdigest(hash_type)
                            for hash_type in StreamHasher.KNOWN_HASH_TYPES
                        }
                        hashes[
                            os.fspath(global_settings.LOCATION_SUBPATH / file_name)
                        ] = referenced_files[file_name].copy()

                        referenced_files[file_name]["chunk_size"] = chunk_size
                        referenced_files[file_name]["path"] = file_name
                        referenced_files[file_name]["size"] = file_size
                        if need_presigned_urls:
                            presigned_urls.append(
                                to_connector.presigned_url(
                                    global_settings.LOCATION_SUBPATH / file_name,
                                    expiration=7 * 24 * 60 * 60,
                                )
                            )
                        stream.seek(0)

                    to_transfer = list(referenced_files.values())
                    from_connector = FakeConnector(
                        {"path": ""},
                        "File descriptors connector",
                        file_streams,
                        hashes,
                    )
                    transfer = Transfer(from_connector, to_connector)
                    transfer.transfer_objects(
                        global_settings.LOCATION_SUBPATH, to_transfer
                    )
                except:
                    logger.exception("Exception uploading data.")
                    self.send_message(client, {"success": False})
                    break
                else:
                    response = {
                        "success": True,
                        "presigned_urls": presigned_urls,
                    }
                    if to_transfer:
                        # Only send referenced files in the 'data' storage.
                        # References to files in other storage (for instance
                        # upload) are not stored in the database.
                        if storage_name == "data":
                            future = asyncio.run_coroutine_threadsafe(
                                self.manager.send_referenced_files(to_transfer),
                                self.loop,
                            )
                            future_response = future.result()
                            # When data object is in state ERROR all responses will
                            # have error status to indicate processing should
                            # finish ASAP.
                            if (
                                future_response.response_status == ResponseStatus.ERROR
                                and future_response.message_data != "OK"
                            ):
                                response = {"success": False}
                    self.send_message(client, response)
                finally:
                    for stream in file_streams.values():
                        stream.close()
        logger.debug("Upload run command stopped.")

    def terminate(self):
        """Stop the uploader thread."""
        self._terminating = True


class ListenerProtocol(BaseProtocol):
    """Listener protocol."""

    def __init__(
        self, communicator: BaseCommunicator, processing_communicator: BaseCommunicator
    ):
        """Initialize."""
        super().__init__(communicator, logger)
        self.processing_communicator = processing_communicator

    async def bootstrap(self):
        """Read boostrap data from the listener and modify the setting.

        :raises RuntimeError: when bootstrop data could not be read.
        """
        # Initialize settings constants by bootstraping.
        response = await self.communicator.bootstrap((DATA_ID, "communication"))
        if response.status != ResponseStatus.OK:
            raise RuntimeError("Error reading bootstrap data.")

        global_settings.initialize_constants(DATA_ID, response.message_data)
        modify_connector_settings()
        # Recreate connectors with received settings.
        connectors.recreate_connectors()
        set_default_storage_connectors()

    async def get_script(self) -> str:
        """Update data status to PR and nhach the script from the listener.

        :raises RuntimeError: when there is error communicating with the
            listener service.
        """
        response = await self.communicator.update_status("PR")
        if response.status == ResponseStatus.ERROR:
            raise RuntimeError("Error changing data status to PR.")

        response = await self.communicator.get_script("")
        if response.response_status == ResponseStatus.ERROR:
            raise RuntimeError("Error while fetching script.")

        return response.message_data

    async def finish(self, return_code: int):
        """Send finish command."""
        await self.communicator.finish({"rc": return_code})

    async def handle_terminate(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Handle terminate command."""
        logger.debug(
            "Received terminate command from liseter. Proxying it to the "
            "processing container."
        )
        response = await self.processing_communicator.send_command(message)
        response.uuid = message.uuid
        return response


class ProcessingProtocol(BaseProtocol):
    """Processing protocol."""

    def __init__(
        self,
        manager: "Manager",
        communicator: BaseCommunicator,
        listener_communicator: BaseCommunicator,
    ):
        """Initialize."""
        super().__init__(communicator, logger)
        self.listener_communicator = listener_communicator
        self.manager = manager
        self.return_code = 1
        self.communicator.heartbeat_handler = self.heartbeat_handler

    async def heartbeat_handler(self, identity: PeerIdentity):
        """Proxy heartbeat messages to the listener."""
        await self.listener_communicator.send_heartbeat()

    async def default_command_handler(
        self, message: Message, identity: PeerIdentity
    ) -> Response:
        """Proxy command to the listener."""
        logger.debug(
            "Proxying command '%s' to the processing container.", message.command_name
        )
        return await self.listener_communicator.send_command(message, identity)

    async def handle_finish(self, message: Message[int], identity: PeerIdentity):
        """Process the finish message."""
        self.return_code = message.message_data
        self.manager.stop()
        return message.respond_ok("")

    async def handle_upload_dirs(
        self, message: Message[list[str]], identity: PeerIdentity
    ) -> Response[str]:
        """Create directories and sent them to the listener.

        This is needed in case empty dirs are referenced.
        """
        subpath = global_settings.LOCATION_SUBPATH
        directories = message.message_data
        referenced_dirs = []
        for directory in directories:
            if storage_connectors := STORAGE_CONNECTOR.get("data"):
                if mounted_connector := storage_connectors[1]:
                    destination_dir = mounted_connector.path / subpath / directory
                    destination_dir.mkdir(parents=True, exist_ok=True)
            referenced_dirs.append({"path": os.path.join(directory, ""), "size": 0})
        response = await self.listener_communicator.referenced_files(referenced_dirs)
        response_method = (
            message.respond_ok
            if response.status == ResponseStatus.OK
            else message.respond_error
        )
        return response_method(response.message_data)

    async def process_script(self, script: str):
        """Send the script to the processing container.

        :raises RuntimeError: when script could not be sent to the processing
            container.
        """
        response = await self.communicator.process_script(script)
        if response.status != ResponseStatus.OK:
            raise RuntimeError("Error sending script to the processing container.")

    async def terminate(self, reason: str = ""):
        """Terminate the processing container."""
        await self.communicator.send_command(Message.command("terminate", reason))


class Manager:
    """Main class.

    Communicate with the listener and with the processing container.
    """

    def __init__(self):
        """Initialize."""
        self.processing_communicator: Optional[BaseCommunicator] = None
        self.listener_communicator: Optional[BaseCommunicator] = None
        self.processing_container_connected = asyncio.Event()
        self._process_script_task: Optional[asyncio.Task] = None
        self._terminating = asyncio.Event()

    async def send_referenced_files(self, referenced_files):
        """Send referenced files to the listener."""
        return await self.listener_communicator.referenced_files(referenced_files)

    def stop(self):
        """Stop the communication container."""
        logger.debug("Manager got stop call.")
        self._terminating.set()

    async def _handle_processing_container_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        """Handle incoming connection from the processing container.

        Python process starts a single connection while Resolwe runtime utils
        starts a new connection for every request.
        """
        logger.debug("Processing container connected")

        # Notify the processing container that the connection is ready.
        writer.write(b"PING\n")
        await writer.drain()
        self.processing_container_connected.set()
        self.processing_communicator = SocketCommunicator(
            reader, writer, "(communication <-> processing)", logger
        )

    async def start_processing_socket(self):
        """Start listening on the processing file socket.

        The socket is used by the processing container to communicatite with
        us.
        """
        await asyncio.start_unix_server(
            self._handle_processing_container_connection, os.fspath(PROCESSING_SOCKET)
        )
        logger.debug(
            "Started listening for processing container on '%s'.", PROCESSING_SOCKET
        )

    async def open_listener_connection(self) -> ZMQCommunicator:
        """Connect to the listener service.

        We are using data id as identity. This implies only one process per
        data object at any given point in time can be running.
        """
        zmq_context = zmq.asyncio.Context.instance()
        zmq_socket = zmq_context.socket(zmq.DEALER)
        zmq_socket.setsockopt(zmq.IDENTITY, str(DATA_ID).encode())
        connect_string = f"{LISTENER_PROTOCOL}://{LISTENER_IP}:{LISTENER_PORT}"
        logger.debug("Opening listener connection to '%s'.", connect_string)
        zmq_socket.connect(connect_string)
        return ZMQCommunicator(zmq_socket, "communication <-> listener", logger)

    async def transfer_missing_data(self):
        """Transfer missing data.

        Log error re-raise exception on failure.

        :raises: RuntimeError on failure.
        """
        try:
            await transfer_data(self.listener_communicator)
        except RuntimeError:
            with suppress(Exception):
                await self.listener_communicator.process_log(
                    {"error": ["Error transfering missing data."]}
                )
            raise

    async def start(self) -> int:
        """Start the main program."""
        try:
            loop = asyncio.get_running_loop()

            # Here we have to:
            # - start the upload thread
            # - connect to the listener service
            # - connect to the processing container
            # When something of the above fails, we have to terminate.

            # Start the upload thread.
            logger.debug("Starting upload thread")
            thread_pool_executor = ThreadPoolExecutor(max_workers=1)
            uploader = Uploader(self, loop)

            uploader_task = loop.run_in_executor(thread_pool_executor, uploader.run)

            # Start server on the socket so processing container can connect.
            await self.start_processing_socket()

            # Open connection to the listener service.
            self.listener_communicator = await self.open_listener_connection()

            # First Wait up to 60 seconds for uploader to get ready.
            if not uploader.ready.wait(60):
                logger.error("Upload thread failed to start, terminating.")
                raise RuntimeError("Upload thread failed to start.")

            # Wait for the processing container to connect to the socket.
            try:
                logger.debug("Waiting for the processing container to connect")
                await asyncio.wait_for(
                    self.processing_container_connected.wait(),
                    constants.CONTAINER_TIMEOUT,
                )
            except asyncio.TimeoutError:
                message = "Unable to connect to the processing container."
                logger.critical(message)
                raise RuntimeError(message)

            logger.debug("Connected to the processing container.")

            listener = ListenerProtocol(
                self.listener_communicator, self.processing_communicator
            )
            processing = ProcessingProtocol(
                self, self.processing_communicator, self.listener_communicator
            )

            try:
                # Start listening for messages from the communication and the
                # processing container.
                listener_task = asyncio.create_task(listener.communicate())
                processing_task = asyncio.create_task(processing.communicate())

                # Read bootstrap data from the listener.
                await listener.bootstrap()

                # Send script to the processing container for processing.
                try:
                    script = await listener.get_script()
                except RuntimeError:
                    logger.exception("Error fetching script")
                    logger.exception("Terminating processing container")
                    await processing.terminate()
                else:
                    await processing.process_script(script)

                # Wait until:
                # - upload thread is alive
                # - listener connection is closed
                # - communication container connection is closed
                # - communication container signals end of the processing
                # - listener signals the end of the processing

                terminating_task = asyncio.create_task(self._terminating.wait())

                monitored_tasks = [
                    listener_task,
                    processing_task,
                    uploader_task,
                    terminating_task,
                ]

                await asyncio.wait(monitored_tasks, return_when=asyncio.FIRST_COMPLETED)
            except RuntimeError as runtime_exception:
                logger.exception("Error processing script.")
                with suppress(Exception):
                    await self.listener_communicator.process_log(
                        {
                            "error": [
                                "Runtime error in communication container: "
                                f"{runtime_exception}."
                            ]
                        }
                    )

        except Exception as e:
            logger.exception("Unhandled exception in the communication container.")
            with suppress(Exception):
                await self.listener_communicator.process_log({"error": [str(e)]})

        finally:
            # Notify listener that the processing is finished.
            try:
                await listener.finish(processing.return_code)
            except RuntimeError:
                logger.exception("Error sending finish command.")
            except:
                logger.exception("Unknown error sending finish command.")

            logger.debug("Signaling upload task to stop.")
            uploader.terminate()
            with suppress(asyncio.CancelledError):
                await asyncio.wait_for(uploader_task, timeout=60)

            if not KEEP_DATA:
                logger.debug("Purging secrets.")
                purge_secrets()

            logger.debug("Stopping the processing and listener communicators.")
            processing.stop_communicate()
            listener.stop_communicate()
            # Wait for up to 10 seconds to close the tasks.
            with suppress(asyncio.TimeoutError):
                await asyncio.wait_for(
                    asyncio.gather(listener_task, processing_task), timeout=60
                )

            # Shut down the thread pool executor.
            logger.debug("Stopping the thread pool executor.")
            thread_pool_executor.shutdown()
            logger.debug("Thread pool executor stopped.")
            return processing.return_code


def set_default_storage_connectors():
    """Set default mounted connector for each known storage."""
    storages = global_settings.SETTINGS["FLOW_STORAGE"]
    for storage_name, storage_settings in storages.items():
        storage_connectors = connectors.for_storage(storage_name)
        default_connector = storage_connectors[0]
        is_mountable = default_connector.name in MOUNTED_CONNECTORS
        default_mounted_connector = default_connector if is_mountable else None
        STORAGE_CONNECTOR[storage_name] = (default_connector, default_mounted_connector)


def modify_connector_settings():
    """Modify mountpoints and add processing and input connectors.

    The path settings on filesystem connectors point to the path on the worker
    node. They have to be remaped to a path inside container and processing and
    input connector settings must be added.
    """
    connector_settings = global_settings.SETTINGS["STORAGE_CONNECTORS"]
    storages = global_settings.SETTINGS["FLOW_STORAGE"]
    connector_storage = {
        connector_name: storage_name
        for storage_name, storage_settings in storages.items()
        for connector_name in storage_settings["connectors"]
    }

    # Point connector path to the correct mountpoint.
    for connector_name in MOUNTED_CONNECTORS:
        storage_name = connector_storage[connector_name]
        connector_settings[connector_name]["config"]["path"] = Path(
            f"/{storage_name}_{connector_name}"
        )


def sig_term_handler(manager_task: asyncio.Task):
    """Gracefully terminate the running process."""
    logger.debug("SIG_INT received, shutting down.")
    manager_task.cancel()


async def start_communication_container():
    """Start the processing manager and set SIGINT handler."""
    manager = Manager()
    manager_task = asyncio.create_task(manager.start())

    asyncio.get_event_loop().add_signal_handler(
        signal.SIGINT, functools.partial(sig_term_handler, manager_task)
    )

    # Wait for the manager task to finish.
    return await manager_task


if __name__ == "__main__":
    sys.exit(asyncio.run(start_communication_container()))



""".. Ignore pydocstyle D400.

==================
Resolwe Test Cases
==================

.. autoclass:: resolwe.test.TestCaseHelpers
    :members:

.. autoclass:: resolwe.test.TransactionTestCase
    :members:

.. autoclass:: resolwe.test.TestCase
    :members:

.. automodule:: resolwe.test.testcases.process

.. automodule:: resolwe.test.testcases.api

"""

import os
import shutil

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.test import SimpleTestCase as DjangoSimpleTestCase
from django.test import TestCase as DjangoTestCase
from django.test import TransactionTestCase as DjangoTransactionTestCase

from resolwe.flow.managers.listener.redis_cache import redis_cache
from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors


class TestCaseHelpers(DjangoSimpleTestCase):
    """Mixin for test case helpers."""

    def _pre_setup(self, *args, **kwargs):
        # NOTE: This is a work-around for Django issue #10827
        # (https://code.djangoproject.com/ticket/10827) that clears the
        # ContentType cache before permissions are setup.
        ContentType.objects.clear_cache()
        super()._pre_setup(*args, **kwargs)

    def _get_testing_directories(self):
        """Get the testing directories."""
        dirs = [connector.path for connector in connectors.for_storage("data")]
        dirs += [connector.path for connector in connectors.for_storage("upload")]
        dirs += [
            storage_settings.FLOW_VOLUMES[volume_name]["config"]["path"]
            for volume_name in ["processing", "input"]
            if volume_name in storage_settings.FLOW_VOLUMES
        ]
        return dirs

    def _clean_up(self):
        """Clean up after test."""
        if not self._keep_data:
            # Do delete this here. See comment below near the makedirs
            # in setUp.
            for directory in self._get_testing_directories():
                shutil.rmtree(directory, ignore_errors=True)

    def setUp(self):
        """Prepare environment for test."""
        super().setUp()

        redis_cache.clear()

        # Directories need to be recreated here in case a previous
        # TestCase deleted them. Moving this logic into the test runner
        # and manager infrastructure would not work, because the manager
        # and listener can't know where the testcase boundaries are,
        # they just see a series of data objects; deleting too soon
        # might cause problems for some tests. The runner does not have
        # any code between tests, so could only remove data at the very
        # end, by which time it's already too late, since some tests may
        # deal specifically with the purging functionality and should
        # start in a clean environment, without the sediment
        # (e.g. jsonout.txt, stdout.txt) from previous tests.
        for directory in self._get_testing_directories():
            os.makedirs(directory, exist_ok=True)
        self._keep_data = settings.FLOW_MANAGER_KEEP_DATA

        self.addCleanup(self._clean_up)

    def keep_data(self, mock_purge=True):
        """Do not delete output files after tests."""
        self.fail(
            "*ERROR* TestCaseHelpers.keep_data() is deprecated and does not work anymore.\n"
            "Using it will result in attribute errors in the future.\n"
            "Please use the command line options --keep-data and --no-mock-purge instead.\n"
        )

    def assertAlmostEqualGeneric(self, actual, expected, msg=None):
        """Assert almost equality for common types of objects.

        This is the same as :meth:`~unittest.TestCase.assertEqual`, but using
        :meth:`~unittest.TestCase.assertAlmostEqual` when floats are encountered
        inside common containers (currently this includes :class:`dict`,
        :class:`list` and :class:`tuple` types).

        :param actual: object to compare
        :param expected: object to compare against
        :param msg: optional message printed on failures
        """
        self.assertEqual(type(actual), type(expected), msg=msg)

        if isinstance(actual, dict):
            self.assertCountEqual(actual.keys(), expected.keys(), msg=msg)
            for key in actual.keys():
                self.assertAlmostEqualGeneric(actual[key], expected[key], msg=msg)
        elif isinstance(actual, (list, tuple)):
            for actual_item, expected_item in zip(actual, expected):
                self.assertAlmostEqualGeneric(actual_item, expected_item, msg=msg)
        elif isinstance(actual, float):
            self.assertAlmostEqual(actual, expected, msg=msg)
        else:
            self.assertEqual(actual, expected, msg=msg)


class TransactionTestCase(TestCaseHelpers, DjangoTransactionTestCase):
    """Base class for writing Resolwe tests not enclosed in a transaction.

    It is based on Django's :class:`~django.test.TransactionTestCase`.
    Use it if you need to access the test's database from another
    thread/process.

    """

    def setUp(self):
        """Initialize test data."""
        super().setUp()

        user_model = get_user_model()
        self.admin = user_model.objects.create_superuser(
            username="admin",
            email="admin@test.com",
            password="admin",
            first_name="James",
            last_name="Smith",
        )
        self.contributor = user_model.objects.create_user(
            username="contributor",
            email="contributor@test.com",
            first_name="Joe",
            last_name="Miller",
        )
        self.user = user_model.objects.create_user(
            username="normal_user",
            email="user@test.com",
            first_name="John",
            last_name="Williams",
        )

        self.group = Group.objects.create(name="Users")
        self.group.user_set.add(self.user)


class TestCase(TransactionTestCase, DjangoTestCase):
    """Base class for writing Resolwe tests.

    It is based on :class:`~resolwe.test.TransactionTestCase` and
    Django's :class:`~django.test.TestCase`.
    The latter encloses the test code in a database transaction that is
    rolled back at the end of the test.

    """



# pylint: disable=missing-docstring
import io
import os
import shutil
import tempfile
from pathlib import Path
from unittest.mock import patch

from django.conf import settings

from resolwe.storage.connectors import (
    AwsS3Connector,
    GoogleConnector,
    LocalFilesystemConnector,
    connectors,
)
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.connectors.hasher import StreamHasher
from resolwe.test import TestCase

CONNECTORS_SETTINGS = {
    "local": {
        "connector": "resolwe.storage.connectors.localconnector.LocalFilesystemConnector",
        "config": {"path": "/my/path"},
    },
    "S3": {
        "connector": "resolwe.storage.connectors.s3connector.AwsS3Connector",
        "config": {
            "bucket": "genialis-test-storage",
            "credentials": os.path.join(
                settings.PROJECT_ROOT, "testing_credentials_s3.json"
            ),
        },
    },
    "GCS": {
        "connector": "resolwe.storage.connectors.googleconnector.GoogleConnector",
        "config": {
            "bucket": "genialis_storage_test",
            "credentials": os.path.join(
                settings.PROJECT_ROOT, "testing_credentials_gcs.json"
            ),
        },
    },
}


def create_file(name, size):
    with open(name, "wb") as f:
        f.seek(size - 1)
        f.write(b"\0")
        f.seek(0)
        f.write(Path(name).name.encode("utf-8"))


def hashes(filename):
    hasher = StreamHasher()
    with open(filename, "rb") as stream:
        hasher.compute(stream)
    return {type_: hasher.hexdigest(type_) for type_ in ["awss3etag", "md5", "crc32c"]}


class RegistryTest(TestCase):
    def setUp(self):
        with patch("resolwe.storage.settings.STORAGE_CONNECTORS", CONNECTORS_SETTINGS):
            connectors.recreate_connectors()

    def tearDown(self):
        connectors.recreate_connectors()

    def test_instance_from_settings(self):
        self.assertIn("local", connectors)
        self.assertIn("S3", connectors)
        self.assertIn("GCS", connectors)
        for connector_name in connectors:
            connector = connectors[connector_name]
            self.assertIsInstance(connector, BaseStorageConnector)


class TestMixin:
    """Tests for all storage connectors."""

    def test_exists(self):
        for file_ in self.prefix_created_files:
            self.assertTrue(self.connector.exists(file_))
        self.assertFalse(self.connector.exists(self.prefix_name("nonexisting_file")))
        with self.assertRaises(ValueError):
            self.connector.exists("/etc")
        with self.assertRaises(TypeError):
            self.connector.exists(0)

    def test_list_objects(self):
        files = self.connector.get_object_list(self.url_prefix)
        self.assertEqual(len(files), len(self.created_files))
        for file_ in self.created_files:
            self.assertIn(file_, files)

        self.assertEqual(
            self.connector.get_object_list(self.prefix_name("dir")),
            ["3"],
        )
        self.assertEqual(
            self.connector.get_object_list(self.prefix_name("nonexisting_directory")),
            [],
        )
        with self.assertRaises(ValueError):
            self.connector.get_object_list("/etc")
        with self.assertRaises(TypeError):
            self.connector.get_object_list(0)

    def test_delete(self):
        self.connector.delete(self.url_prefix, ["nonexisting_file"])
        files = self.connector.get_object_list(self.url_prefix)
        for file_ in self.created_files:
            self.assertIn(file_, files)
        self.connector.delete(self.url_prefix, self.created_files[:2])
        files = self.connector.get_object_list(self.url_prefix)
        self.assertEqual(files, [self.created_files[2]])
        with self.assertRaises(ValueError):
            self.connector.delete("/etc", [])
        with self.assertRaises(TypeError):
            self.connector.delete(0, [])
        with self.assertRaises(TypeError):
            self.connector.delete("", [0])
        with self.assertRaises(ValueError):
            self.connector.delete("", ["OK", "/etc"])

    def test_get(self):
        stream = io.BytesIO()
        starts_with = Path(self.path(self.prefix_created_files[0])).name.encode("utf-8")
        self.connector.get(self.prefix_created_files[0], stream)
        self.assertEqual(stream.tell(), 1024 * 1024)
        stream.seek(0)
        self.assertEqual(stream.read(len(starts_with)), starts_with)
        self.assertEqual(stream.read(), b"\0" * (1024 * 1024 - len(starts_with)))
        with self.assertRaises(ValueError):
            self.connector.get("/etc", None)
        with self.assertRaises(TypeError):
            self.connector.get(0, None)

    def test_push(self):
        transfer_file = self.prefix_name("transfer")
        with open(self.path(self.prefix_created_files[0]), "rb") as f:
            self.connector.push(f, transfer_file)
        self.assertTrue(self.connector.exists(transfer_file))

        downloaded_file = self.prefix_name("download")
        with open(self.path(downloaded_file), "wb") as stream:
            self.connector.get(transfer_file, stream)

        with open(self.path(self.prefix_created_files[0]), "rb") as f1:
            with open(self.path(downloaded_file), "rb") as f2:
                self.assertEqual(f1.read(), f2.read())

        with self.assertRaises(ValueError):
            self.connector.push(None, "/etc")
        with self.assertRaises(TypeError):
            self.connector.push(None, 0)

        self.connector.delete(self.url_prefix, ["transfer"])

    def test_get_hash(self):
        test_file = self.prefix_created_files[0]
        computed_hashes = hashes(self.path(test_file))
        for type_ in ("md5", "crc32c", "awss3etag"):
            self.assertEqual(
                self.connector.get_hash(test_file, type_).lower(),
                computed_hashes[type_].lower(),
            )
        with self.assertRaises(ValueError):
            self.connector.get_hash("/etc", None)
        with self.assertRaises(TypeError):
            self.connector.get_hash(0, None)


class BaseTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        setattr(cls, "tmp_dir", tempfile.TemporaryDirectory())
        local_settings = CONNECTORS_SETTINGS["local"]["config"].copy()
        local_settings["path"] = cls.tmp_dir.name
        cls.local = LocalFilesystemConnector(local_settings, "local")
        # Afects awss3etag hash computation.
        cls.local.multipart_chunksize = 8 * 1024 * 1024
        cls.gcs = GoogleConnector(CONNECTORS_SETTINGS["GCS"]["config"], "GCS")
        cls.s3 = AwsS3Connector(CONNECTORS_SETTINGS["S3"]["config"], "S3")

    def _purge_tmp_dir(self):
        for root, dirs, files in os.walk(self.tmp_dir.name):
            for f in files:
                os.unlink(os.path.join(root, f))
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))

    def setUp(self):
        self._purge_tmp_dir()
        os.makedirs(os.path.join(self.tmp_dir.name, self.url_prefix, "dir"))

        self.prefix_created_files = [
            os.path.join(self.url_prefix, name) for name in ("1", "2", "dir/3")
        ]
        self.created_files = [name for name in ("1", "2", "dir/3")]

        for file_ in self.prefix_created_files:
            create_file(self.path(file_), 1024 * 1024)

        # Delete possible transfer file.
        self.connector.delete(self.url_prefix, ["transfer"])

        return super().setUp()

    def path(self, filename):
        return os.path.join(self.tmp_dir.name, filename)

    def prefix_name(self, filename):
        return os.path.join(self.url_prefix, filename)


class LocalConnectorTest(BaseTestCase, TestMixin):
    def setUp(self):
        self.url_prefix = "local_connector_test"
        self.connector = self.local
        super().setUp()


class GoogleConnectorTest(BaseTestCase, TestMixin):
    def setUp(self):
        self.url_prefix = "google_connector_test"
        self.connector = self.gcs
        super().setUp()
        # Use connector to push files to storage for testing.
        # If this does not work tests will fail anyway.
        for file_ in self.prefix_created_files:
            if not self.gcs.exists(file_):
                with open(self.path(file_), "rb") as stream:
                    self.gcs.push(stream, file_)
                    computed_hashes = hashes(self.path(file_))
                    self.gcs.set_hashes(file_, computed_hashes)


class S3ConnectorTest(BaseTestCase, TestMixin):
    def setUp(self):
        self.url_prefix = "s3_connector_test"
        self.connector = self.s3
        super().setUp()
        # Use connector to push files to storage for testing.
        # If this does not work tests will fail anyway.
        for file_ in self.prefix_created_files:
            with open(self.path(file_), "rb") as stream:
                if not self.s3.exists(file_):
                    self.s3.push(stream, file_)
                    computed_hashes = hashes(self.path(file_))
                    self.s3.set_hashes(file_, computed_hashes)



"""Custom permissions for Flow API."""
from typing import Optional

from django.contrib.auth.models import User
from django.db import models
from django.http import Http404, HttpRequest

from rest_framework import exceptions, permissions

from resolwe.permissions.models import Permission, PermissionList, get_anonymous_user
from resolwe.permissions.utils import get_user, model_has_permissions


class ResolwePermissionBackend:
    """Custom resolwe permission backend."""

    supports_object_permissions = True
    supports_anonymous_user = True
    supports_inactive_user = True

    def authenticate(
        self,
        request: HttpRequest,
        username: Optional[User] = None,
        password: Optional[str] = None,
    ):
        """Authenticate user.

        Since our backend is only used for authorization return None. See

        https://docs.djangoproject.com/en/3.2/topics/auth/customizing/

        for details.
        """
        return None

    def has_perm(
        self, user: User, permission: Permission, obj: Optional[models.Model] = None
    ) -> bool:
        """
        Check if user has the given permission on the given object.

        :returns: ``True`` if given ``user_obj`` has ``perm`` for ``obj``.
            If no ``obj`` is given or user is inactive and authenticated at the
            same time ``False`` is returned.
        """
        user = get_user(user)
        if not user.is_active:
            return False

        if not model_has_permissions(obj):
            return True

        return obj.has_permission(permission, user)

    def get_all_permissions(
        self, user: User, obj: Optional[models.Model] = None
    ) -> PermissionList:
        """Return a set of permissions for ``user`` on ``obj``.

        Permissions from user groups are also considered.

        When user is not active the empty set of permissions is returned.
        """
        all_permissions = []
        if user.is_active and model_has_permissions(obj):
            permission_value = obj.permission_group.permissions.filter(
                models.Q(user=user) | models.Q(group__in=user.groups.all())
            ).aggregate(permission=models.Max("value"))["permission"]
            if permission_value:
                all_permissions = list(Permission(permission_value))
        return all_permissions


class ResolwePermissions(permissions.DjangoObjectPermissions):
    """Resolwe Permissions."""

    perms_map = {
        "GET": [Permission.VIEW],
        "OPTIONS": [Permission.VIEW],
        "HEAD": [Permission.VIEW],
        "POST": [Permission.EDIT],
        "PUT": [Permission.EDIT],
        "PATCH": [Permission.EDIT],
        "DELETE": [Permission.EDIT],
    }

    def get_required_object_permissions(self, method, model_cls):
        """Return the list of permissions that the user is required to have."""
        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return self.perms_map[method]

    def has_permission(self, request, view):
        """Return `True` as we don't use model level perms."""
        return True

    def has_object_permission(self, request, view, obj):
        """Check object permissions."""

        # admins can do anything
        if request.user.is_superuser:
            return True

        # No permission on objects that do not support it.
        if not model_has_permissions(obj):
            return False

        # `share` permission is required for editing permissions
        if view.action is not None and "permissions" in view.action:
            self.perms_map["POST"] = [Permission.SHARE]

        if hasattr(view, "get_queryset"):
            queryset = view.get_queryset()
        else:
            queryset = getattr(view, "queryset", None)

        assert queryset is not None, (
            "Cannot apply DjangoObjectPermissions on a view that "
            "does not set `.queryset` or have a `.get_queryset()` method."
        )

        model_cls = queryset.model
        user = request.user
        perm = max(self.get_required_object_permissions(request.method, model_cls))
        anonymous = get_anonymous_user()
        if not user.has_perm(perm, obj) and not anonymous.has_perm(perm, obj):
            if request.method not in permissions.SAFE_METHODS:
                # If the user does not have permissions we need to determine if
                # they have read permissions to see 403, or not, and simply see
                # a 404 response.
                read_perm = max(self.get_required_object_permissions("GET", model_cls))

                if user.has_perm(read_perm, obj):
                    return False

            raise Http404

        return True
